/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

package uk.co.petertribble.jangle;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Date;
import java.awt.event.*;
import java.math.BigInteger;

import org.jfree.chart.*;
import org.jfree.data.time.*;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.XYPlot;

/**
 * Display snmp data in a graphical chart.
 *
 * @author Peter Tribble
 */
public class SnmpChart implements ActionListener {

    static private int DEFAULT_INTERVAL = 30;
    static private SnmpMibManager smm;

    private SnmpController sc;
    private JFreeChart chart;
    private javax.swing.Timer timer;
    private int interval;
    private int maxage = 1800000;
    private TimeSeriesCollection dataset;
    private boolean showdelta;
    // times in milliseonds
    private long lastsnap;
    private String charttitle;

    private Map <String, TimeSeries> tsmap;
    // save previous values for rates
    private Map <String, BigInteger> valueMap;

    static {
	smm = SnmpMibManager.getInstance();
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given oid.
     *
     * @param sc an SnmpController
     * @param oid the oid to show
     * @param charttitle the title to show on the chart
     */
    public SnmpChart(SnmpController sc, String oid, String charttitle) {
	this(sc, oid, charttitle, true, DEFAULT_INTERVAL);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given oid.
     *
     * @param sc an SnmpController
     * @param oid the oid to show
     * @param charttitle the title to show on the chart
     * @param interval the update interval in seconds
     */
    public SnmpChart(SnmpController sc, String oid, String charttitle,
		int interval) {
	this(sc, oid, charttitle, true, interval);
    }

    /**
     * Create a new SnmpChart, showing the given oid.
     *
     * @param sc an SnmpController
     * @param oid the oid to show
     * @param charttitle the title to show on the chart
     * @param showdelta if true, show rates instead of values
     */
    public SnmpChart(SnmpController sc, String oid, String charttitle,
		boolean showdelta) {
	this(sc, oid, charttitle, showdelta, DEFAULT_INTERVAL);
    }

    /**
     * Create a new SnmpChart, showing the given oid.
     *
     * @param sc an SnmpController
     * @param oid the oid to show
     * @param charttitle the title to show on the chart
     * @param showdelta if true, show rates instead of values
     * @param interval the update interval in seconds
     */
    public SnmpChart(SnmpController sc, String oid, String charttitle,
		boolean showdelta, int interval) {
	this.sc = sc;
	this.showdelta = showdelta;
	this.charttitle = charttitle;
	this.interval = interval;
	initialize(oid);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given OIDs.
     *
     * @param sc an SnmpController
     * @param oids a List of OIDs to show
     * @param charttitle the title to show on the chart
     */
    public SnmpChart(SnmpController sc, List <SnmpObject> oids,
		String charttitle) {
	this(sc, oids, charttitle, true, DEFAULT_INTERVAL);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given OIDs.
     *
     * @param sc an SnmpController
     * @param oids a List of oids to show
     * @param charttitle the title to show on the chart
     * @param interval the update interval in seconds
     */
    public SnmpChart(SnmpController sc, List <SnmpObject> oids,
		String charttitle, int interval) {
	this(sc, oids, charttitle, true, interval);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given OIDs.
     *
     * @param sc an SnmpController
     * @param oids a List of oids to show
     * @param charttitle the title to show on the chart
     * @param showdelta if true, show rates instead of values
     */
    public SnmpChart(SnmpController sc, List <SnmpObject> oids,
		String charttitle, boolean showdelta) {
	this(sc, oids, charttitle, showdelta, DEFAULT_INTERVAL);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given OIDs.
     *
     * @param sc an SnmpController
     * @param snos a List of oids to show
     * @param charttitle the title to show on the chart
     * @param showdelta if true, show rates instead of values
     * @param interval the update interval in seconds
     */
    public SnmpChart(SnmpController sc, List <SnmpObject> snos,
		String charttitle, boolean showdelta, int interval) {
	this.sc = sc;
	this.showdelta = showdelta;
	this.charttitle = charttitle;
	this.interval = interval;
	List <String> oids = new ArrayList <String> ();
	for (SnmpObject sno : snos) {
	    oids.add(sno.toString());
	}
	initialize(oids);
    }

    private void initialize(String oid) {
	List <String> oids = new ArrayList <String> ();
	oids.add(oid);
	initialize(oids);
    }

    private void initialize(List <String> oids) {
	tsmap = new HashMap <String, TimeSeries> ();
	valueMap = new HashMap <String, BigInteger> ();
	dataset = new TimeSeriesCollection();
	lastsnap = 0;

	for (String oid : oids) {
	    TimeSeries ts = new TimeSeries(smm.prettifyOID(oid));
	    ts.setMaximumItemAge(maxage);
	    dataset.addSeries(ts);
	    tsmap.put(oid, ts);
	    valueMap.put(oid, BigInteger.ZERO);
	}

	updateAccessory();

	String ylabel = showdelta ? SnmpResources.getString("CHART.RATE")
	    : SnmpResources.getString("CHART.VALUE");

	chart = ChartFactory.createTimeSeriesChart(
		charttitle,
		SnmpResources.getString("CHART.TIME"),
		ylabel,
		dataset,
		true,
		true,
		false);

	setAxes();

	startLoop();
    }

    /*
     * Set up the X and Y axes.
     */
    private void setAxes() {
	XYPlot xyplot = chart.getXYPlot();

	String ylabel = showdelta ? SnmpResources.getString("CHART.RATE")
	    : SnmpResources.getString("CHART.VALUE");
	NumberAxis loadaxis = new NumberAxis(ylabel);
	loadaxis.setAutoRange(true);
	loadaxis.setAutoRangeIncludesZero(true);
	xyplot.setRangeAxis(loadaxis);

	DateAxis daxis = new DateAxis(SnmpResources.getString("CHART.TIME"));
	daxis.setAutoRange(true);
	daxis.setFixedAutoRange(maxage);
	xyplot.setDomainAxis(daxis);
    }

    /**
     * Return the chart that is created.
     *
     * @return  The created chart
     */
    public JFreeChart getChart() {
	return chart;
    }

    /**
     * Add another oid to the current chart. If the oid is already being
     * charted, it is not added.
     *
     * @param oid the name of the oid to add
     */
    public void addOid(String oid) {
	if (!tsmap.containsKey(oid)) {
	    TimeSeries ts = new TimeSeries(oid);
	    ts.setMaximumItemAge(maxage);
	    dataset.addSeries(ts);
	    tsmap.put(oid, ts);
	    /*
	     * Save the current value, not zero. This ensures that the first
	     * rate calculation gives sensible values. It will be low, as it
	     * will average the change between now and the next update over
	     * the whole update interval. This is better than nothing, and
	     * the way to do this properly would be to have separate lastsnap
	     * values saved for each oid. This would be overkill when
	     * the only affected datapoint is the first one.
	     */
	    // FIXME get the actual value
	    valueMap.put(oid, BigInteger.ZERO);
	}
    }

    /**
     * Remove the requested oid from the current chart.
     *
     * @param oid the name of the oid to remove
     */
    // FIXME is it valid to remove all oids?
    public void removeOid(String oid) {
	if (tsmap.containsKey(oid)) {
	    dataset.removeSeries(tsmap.get(oid));
	    tsmap.remove(oid);
	    valueMap.remove(oid);
	}
    }

    /**
     * Set the maximum age of the chart. Only oids younger than this age will
     * be shown.
     *
     * @param age The required maximum age in seconds.
     */
    public void setMaxAge(int age) {
	maxage = age*1000;
	for (TimeSeries ts : tsmap.values()) {
	    ts.setMaximumItemAge(maxage);
	}
    }

    /**
     * Update the oids.
     */
    public void updateAccessory() {
	double value;
	long newsnap = (new Date()).getTime();
	// loop over all oids
	for (String stat : tsmap.keySet()) {
	    try {
		BigInteger newvalue = sc.getValue(stat).getNumber();
		if (showdelta) {
		    BigInteger bd = newvalue.subtract(valueMap.get(stat));
		    double dt = (double) (newsnap - lastsnap);
		    value = 1000.0*(bd.doubleValue()/dt);
		    valueMap.put(stat, newvalue);
		} else {
		    value = newvalue.doubleValue();
		}
		tsmap.get(stat).add(new Millisecond(), value);
	    } catch (SnmpException sne) {}
	}
	lastsnap = newsnap;
    }

    /**
     * Start the timer, so the display will continually update.
     */
    public void startLoop() {
	if (timer == null) {
	    timer = new javax.swing.Timer(interval*1000, this);
	}
	timer.start();
    }

    /**
     * Stop the timer, so the display will not update.
     */
    public void stopLoop() {
	if (timer != null) {
	    timer.stop();
	}
    }

    /**
     * Set the update delay.
     *
     * @param interval the delay value in seconds
     */
    public void setDelay(int interval) {
	this.interval = interval;
	if (timer != null) {
	    timer.setDelay(interval*1000);
	}
    }

    public void actionPerformed(ActionEvent e) {
	updateAccessory();
    }
}
